import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib as mpl


class MplColorHelper:
    def __init__(self, cmap_name, cmap_lim, centered_norm=True):
        self.cmap_name = cmap_name
        self.cmap = plt.get_cmap(cmap_name)
        if centered_norm:
            self.norm = mpl.colors.CenteredNorm(
                vcenter=cmap_lim[0], halfrange=cmap_lim[1], clip=False
            )
        else:
            self.norm = mpl.colors.Normalize(vmin=cmap_lim[0], vmax=cmap_lim[1])
        self.scalarMap = cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_rgb(self, val, alpha):
        return self.scalarMap.to_rgba(val, alpha=alpha)



import warnings
from itertools import permutations
from typing import NoReturn, Optional, Tuple, Callable, Union, Generator

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from ._aggregator import get_batch_creator
from ._masking import IndependentMaskedModel


class Shapley:
    """Compute the Shapley value for every element of a collection of instances that are aggregated together in a single
    input of a trained predictive algorithm. It either uses an exact enumeration strategy when the number of samples is
    relatively small (<8) or an approximate Monte-Carlo scheme with antithetic sampling.

    Parameters
    ----------
    predictor: callable (input: 2D array of shape (n_inputs, n_input_features), output: 1D array of shape (n_inputs,))
        Trained predictor that returns one single real-value prediction per input.

    aggregation: callable, tuple, list of tuples
        Aggregator that transforms an array of n_instances (with each one characterized by n_instance_features) into one
        single vector of shape (n_input_features) that will be used as an input of the predictor.

        To define the aggregator one can use:

            * a callable that takes as input a 2D array of shape (n_instances, n_instance_features) and returns a 1D array of shape (1, n_input_features).
            * a tuple (`method`, `subset`) with `method` being a string that refers to a numpy aggregating function (e.g 'sum', 'min', 'std'...) and `subset` being a 1D array that defines the subset of columns/features on which to apply this method (or None for applying it on all the columns/features).
            * a list of tuples [(`method_1`, `subset_1`), (`method_2`, `subset_2`), ...] to define several aggregators. Please note that the aggregated features will be ordered according to the order of the provided list (i.e [agg_feature_method_1, ..., agg_feature_method_2, ...]).

    empty_value: float, optional
        Prediction made by the algorithm for an input with no instances (i.e random prediction). The default is 0.5.

    Attributes
    ----------
    shapleyvalues_: 1D array of shape (n_instances,)
        Shapley value associated to each instance.

    Notes
    -----
    For an exact estimation, the sum of the Shapley values of all instances equals the difference between the prediction
    made for the collection of all instances and the empty/random prediction (Efficiency property).

    Examples
    --------
    >>> import numpy as np
    >>> import joblib
    >>> from radshap.shapley import Shapley
    >>>
    >>> model = joblib.load("trained_logistic_regression.joblib")
    >>> shap = Shapley(predictor = lambda x: model.predict_proba(x)[:, 1], aggregation = ('mean', None))
    >>> shapvalues = shap.explain(X) # X a 2D array of shape (n_instances, n_instance_features)
    """

    def __init__(
            self,
            predictor: Callable[[np.ndarray], float],
            aggregation: Callable[[np.ndarray], np.ndarray],
            empty_value: Optional[float] = 0.5,
    ) -> NoReturn:

        self.predictor = predictor
        self.empty_value = empty_value

        self.batch_creator = get_batch_creator(aggregation)
        self._ninstances = None
        self.shapleyvalues_ = None

    def explain(
            self,
            X: np.ndarray,
            estimation_method: Optional[str] = "auto",
            nsamples: Optional[int] = 1000,
            n_jobs: Optional[int] = 1,
    ) -> np.ndarray:
        """Compute the Shapley values for each row of X.

        X must correspond to a valid collection of instances (shape (n_instances, n_instance_features)) that will be
        passed to the aggregator function and then used as an input for the predictor.

        Parameters
        ----------
        X: 2D array of shape (n_instances, n_instance_features)
            X corresponds to a collection of instances that will be aggregated into a single input.

        estimation_method: str {'auto', 'exact', 'antithetic'}, optional
            Estimation method for the Shapley values. The default is "auto".

            * If `estimation = 'exact'`, an exact enumeration strategy is used. All the permutations of the instances are considered.
            * If `estimation = 'antithetic'`, a Monte-Carlo scheme with random permutations is applied. Antithetic sampling is used to reduce the variance of the estimator. In that case the number of samples is defined by `nsamples`.
            * If `estimation = 'auto'`, the estimation method is chosen based on the number of instances. If the number of instances is > 8 a Monte-Carlo scheme is applied. Otherwise, an exact scheme is applied.

        nsamples: str, optional
            Number of samples for the Monte-Carlo scheme. It is not used only when`estimation = 'exact' or when the
            number of instances is > 8. The default is 1000.

        n_jobs: int, optional
            Number of jobs to run in parallel. -1 means using all processors.
            See the joblib package documentation for more explanations. The default is 1.

        Returns
        -------
        1D array of shape (n_instances,)
            Shapley value associated to each instance.
        """
        self._ninstances = X.shape[0]
        self.batch_creator.check_fun_agg(X)

        if estimation_method == "auto":
            if self._ninstances <= 7:
                return self._explain_exact(X, n_jobs)
            else:
                return self._explain_antithetic(X, nsamples, n_jobs)
        elif estimation_method == "exact":
            if self._ninstances > 7:
                warnings.warn(
                    "The number of permutations to test for exact estimation is at least 40320."
                    " It may require some computational power."
                )
            return self._explain_exact(X, n_jobs)
        elif estimation_method == "antithetic":
            return self._explain_antithetic(X, nsamples, n_jobs)
        else:
            raise ValueError(
                "Unrecognized estimation method. Please choose among 'auto', 'exact', or antithetic'"
            )

    def _explain_antithetic(
            self, X: np.ndarray, nsamples: int, n_jobs: int
    ) -> np.ndarray:
        """Apply Monte-Carlo scheme with antithetic sampling for estimating the Shapley values"""
        results = np.zeros((nsamples // 2, self._ninstances))
        parallel = Parallel(n_jobs=n_jobs, verbose=0)
        sampling_results = parallel(
            delayed(self._get_antithetic_evaluations)(X) for _ in range(nsamples // 2)
        )

        for count, value in enumerate(sampling_results):
            results[count, value[0]] = value[1]

        self.shapleyvalues_ = (1 / nsamples) * np.sum(results, axis=0)
        return self.shapleyvalues_

    def _get_antithetic_evaluations(self, X: np.ndarray) -> object:
        """Compute one antithetic sample"""
        p = np.random.permutation(np.arange(self._ninstances))
        marginals = _get_evaluations(
            p, self.batch_creator, self.predictor, self.empty_value, X
        )[1]

        p_r = np.flip(p)
        marginals_r = np.flip(
            _get_evaluations(
                p_r, self.batch_creator, self.predictor, self.empty_value, X
            )[1]
        )
        return p, marginals + marginals_r

    def _explain_exact(self, X: np.ndarray, n_jobs: int) -> np.ndarray:
        """Apply exact enumeration (i.e. run through all the possible permutations for estimating the
        Shapley values)"""
        n = np.math.factorial(self._ninstances)
        results = np.zeros((n, self._ninstances))
        parallel = Parallel(n_jobs=n_jobs, verbose=0)
        exact_results = parallel(
            delayed(_get_evaluations)(
                perm=p,
                batch_creator=self.batch_creator,
                predictor=self.predictor,
                empty_value=self.empty_value,
                X=X,
            )
            for p in _get_permutations(self._ninstances)
        )

        for count, value in enumerate(exact_results):
            results[count, value[0]] = value[1]

        self.shapleyvalues_ = (1 / n) * np.sum(results, axis=0)
        return self.shapleyvalues_


class RobustShapley(Shapley):
    """ Compute the Shapley value while dealing with invalid inputs for every element of a collection of instances that
    are aggregated together in a single input of a trained predictive algorithm.

    Deal with the cases where the aggregation function returns invalid inputs for the predictive algorithm
    (i.e. unexpected NaN values) by replacing missing values by values from a background data set and computing the
    average prediction over all these replacements.

    Parameters
    ----------
    predictor: callable (input: 2D array of shape (n_inputs, n_input_features), output: 1D array of shape (n_inputs,))
        Trained predictor that returns one single real-value prediction per input.

    aggregation: callable, tuple, list of tuples
        Aggregator that transforms an array of n_instances (with each one characterized by n_instance_features) into one
        single vector of shape (n_input_features) that will be used as an input of the predictor.

        Aggregator should deal with cases where some aggregated features cannot be computed by returning NaN values
        instead.

        To define the aggregator one can use:

            * a callable that takes as input a 2D array of shape (n_instances, n_instance_features) and returns a 1D array of shape (1, n_input_features).
            * a tuple (`method`, `subset`) with `method` being a string that refers to a numpy aggregating function (e.g 'sum', 'min', 'std'...) and `subset` being a 1D array that defines the subset of columns/features on which to apply this method (or None for applying it on all the columns/features).
            * a list of tuples [(`method_1`, `subset_1`), (`method_2`, `subset_2`), ...] to define several aggregators. Please note that the aggregated features will be ordered according to the order of the provided list (i.e [agg_feature_method_1, ..., agg_feature_method_2, ...]).

    background_data: 2D array of shape (n_background, n_input_features)
        Background data set used to deal with invalid inputs for the predictor. In the case of an invalid input, we
        generate "n_background" new inputs by replacing the invalid values with corresponding values from the background
        data set. We then apply the predictor to each of these generated inputs and return the average predictions.

    invalid_features: 1D array of shape (n_input_features)
        Boolean array to specify the features for which a NaN value means that the input is not valid. In such cases, we
        handle them with the background data set. If None all the features will be taken into account to define invalid
        inputs. The default is None.

    empty_value: float, optional
        Prediction made by the algorithm for an input with no instances (i.e random prediction). The default is 0.5.

    Attributes
    ----------
    shapleyvalues_: 1D array of shape (n_instances,)
        Shapley value associated to each instance.

    Notes
    -----
    The strategy to deal with invalid inputs was inspired by SHAP (https://shap.readthedocs.io).

    Examples
    --------
    >>> import numpy as np
    >>> import joblib
    >>> from radshap.shapley import RobustShapley
    >>>
    >>> model = joblib.load("trained_logistic_regression.joblib")
    >>> shap = RobustShapley(predictor = lambda x: model.predict_proba(x)[:, 1],
    >>>                      aggregation = ('nanmean', None),
    >>>                      background_data = Xback) # Xback a 2D array of shape (n_samples_background, n_input_features)
    >>> shapvalues = shap.explain(X) # X a 2D array of shape (n_instances, n_instance_features)
    """

    def __init__(self,
                 predictor: Callable[[np.ndarray], float],
                 aggregation: Callable[[np.ndarray], np.ndarray],
                 background_data: np.ndarray,
                 invalid_features: Optional[Union[np.ndarray, None]] = None,
                 empty_value: Optional[float] = 0.5,
                 ) -> NoReturn:
        super(RobustShapley, self).__init__(predictor=IndependentMaskedModel(model=predictor,
                                                                             background_data=background_data,
                                                                             invalid_features=invalid_features),
                                            aggregation=aggregation,
                                            empty_value=empty_value)
        self._background_data = background_data
        self._masked_features = invalid_features


def _get_permutations(n: int) -> Generator[np.ndarray, None, None]:
    """Yield permutations of the natural numbers [0, ..., n-1]"""
    for p in permutations(range(n)):
        yield np.array(p)


def _get_evaluations(
        perm: np.ndarray,
        batch_creator: Callable,
        predictor,
        empty_value,
        X,
) -> Tuple[np.ndarray, np.ndarray]:
    """ """
    batch = batch_creator(perm, X)
    values = predictor(batch)
    marginals = _get_marginals(values, empty_value)
    return perm, marginals


def _get_marginals(v: np.ndarray, first_value: float) -> np.ndarray:
    """ """
    v_ = np.zeros(v.shape)
    v_[0] = first_value
    v_[1:] = v.copy()[0:-1]
    return v - v_



from typing import NoReturn, Optional, Tuple, Callable, Union
import numpy as np

_LIST_NUMPY = [
    "min",
    "max",
    "mean",
    "prod",
    "std",
    "sum",
    "var",
    "nanmin",
    "nanmax",
    "nanmean",
    "nanprod",
    "nanstd",
    "nansum",
    "nanvar",
]


def get_batch_creator(aggregation: Union[tuple, list, Callable]) -> Callable:
    """Build a batch creator that will aggregate several instances in a cumulative way.

    For a set of instances | instance_1 | the batch creator will output |            aggregator(instance_1)       |
                           | instance_2 |                               |     aggregator(instance_1, instance_2)  |
                                ...                                                            ...
                           | instance_k |                               | aggregator(instance_1, ..., instance_k) |
    Parameters
    ----------
    aggregation: tuple, list or tuple or callable
        Aggregation function.

        To define the aggregation funcion one can use::

            * a callable that takes as input a 2D array of shape (n_instances, n_instance_features) and returns a 1D
            array of shape (1, n_input_features).
            * a tuple (`method`, `subset`) with `method` being a string that refers to a numpy aggregating function (e.g
            'sum', 'min', 'std'...) and `subset` being a 1D array that defines the subset of columns/features on which
            to apply this method (or None for applying it on all the columns/features).
            * a list of tuples [(`method_1`, `subset_1`), (`method_2`, `subset_2`), ...] to define several aggregators.
            Please note that the aggregated features will be ordered according to the order of the provided list (i.e
            [agg_feature_method_1, ..., agg_feature_method_2, ...]).

    Returns
    -------
    callable
        Batch creator associated with the aggregation function

    Note
    ----
    If the aggregation function is defined with numpy aggregating functions (i.e. tuple or list of tuples) the batch
    creator will take advantage of numpy operations to speed up computations. Otherwise, a slow for loop will be used.
    """
    if callable(aggregation):
        return _NaiveBatchCreator(aggregation)
    elif isinstance(aggregation, (list, tuple)):
        return _FastBatchCreator(aggregation)
    raise TypeError(
        "aggregation should either be a callable (custom function), a tuple of the form (method (str),"
        "subset (None or np.ndarray)), or a list of such tuples"
    )


class _BatchCreator:
    def __init__(self, aggregation):
        self.fun_agg = _get_aggregation_function(aggregation)

    def check_fun_agg(self, X):
        pass

    def _create(self, permutation: np.ndarray, X: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __call__(self, permutation: np.ndarray, X: np.ndarray) -> np.ndarray:
        return self._create(permutation, X)


class _NaiveBatchCreator(_BatchCreator):
    """Naive Batch creator for generic aggregation functions"""

    def check_fun_agg(self, X):
        try:
            test_results = self.fun_agg(X)
        except Exception as err:
            print(
                "Your custom aggregator is not properly defined."
            )
            print(f"The following {err}, {type(err)} occured")
            raise
        else:
            if (not isinstance(test_results, np.ndarray)) or (
                len(np.squeeze(test_results).shape) > 1
            ):
                print(
                    "Your custom aggregator should take as input a 2D array of shape "
                    "(n_instances, n_instance_features) and it should return a vector of shape (1, n_input_features)"
                    " (or (n_input_features,))."
                )
                raise

    def _create(self, permutation: np.ndarray, X: np.ndarray) -> np.ndarray:

        X_perm = np.copy(X)[permutation, :]
        L = [
            self.fun_agg(X_perm[:i, :]).reshape(1, -1)
            for i in range(1, len(permutation) + 1)
        ]
        return np.vstack(L)


class _FastBatchCreator(_BatchCreator):
    """Fast Batch creator for aggregation functions defined with numpy operations"""

    def _create(self, permutation: np.ndarray, X: np.ndarray) -> np.ndarray:
        n_instances = len(permutation)
        X_perm_3d = np.tile(np.copy(X)[permutation, :], (n_instances, 1, 1))
        mask_3d = np.tile(
            np.triu(np.ones((n_instances, n_instances)), k=1)[:, :, np.newaxis],
            (1, 1, X.shape[1]),
        )
        masked_array = np.ma.array(data=X_perm_3d, mask=mask_3d)
        return self.fun_agg(masked_array)


class _Aggregator3d:
    def __init__(self, method, subset):
        self.method = method
        self.subset = subset

    def __call__(self, masked_3d_array: np.ma.MaskedArray) -> np.ndarray:
        if self.subset is not None:
            return getattr(np, self.method)(
                masked_3d_array[:, :, self.subset], axis=1
            ).data
        return getattr(np, self.method)(masked_3d_array, axis=1).data


def _get_aggregation_function(aggregation: Union[tuple, list, Callable]) -> Callable:
    if callable(aggregation):
        return aggregation
    elif isinstance(aggregation, tuple):
        method, subset = _check_aggregation(aggregation)
        return _Aggregator3d(method=method, subset=subset)
    elif isinstance(aggregation, list):
        aggregation = [_check_aggregation(a) for a in aggregation]
        return lambda m: np.hstack(
            [_Aggregator3d(method=a[0], subset=a[1])(m) for a in aggregation]
        )
    raise TypeError(
        "aggregation should either be a callable (custom function), a tuple of the form (method (str),"
        "subset (None or np.ndarray)), or a list of such tuples"
    )


def _check_aggregation(tuple_aggregation: tuple) -> tuple:
    if (
        (len(tuple_aggregation) != 2)
        or (not isinstance(tuple_aggregation[0], str))
        or (
            not (
                (tuple_aggregation[1] is None)
                or (isinstance(tuple_aggregation[1], (list, np.ndarray)))
            )
        )
    ):
        raise ValueError(
            "aggregation only accepts tuples of the form (method (str), subset (None or np.ndarray))"
        )
    if not tuple_aggregation[0] in _LIST_NUMPY:
        raise ValueError("")
    return tuple_aggregation



from typing import NoReturn, Optional, Tuple, Callable, Union, Generator

import SimpleITK as sitk
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from ._utils import MplColorHelper


def plot_bars(
    shapley_values: np.ndarray,
    nbest: Optional[int] = 10,
    names: Optional[Union[list, None]] = None,
    sort: Optional[bool] = True,
    ax: Optional[Union[matplotlib.axes.Axes, None]] = None,
) -> None:
    """Plot the Shapley values of different instances with a bar plot.

    Parameters
    ----------
    shapley_values : 1D array shape (n_shapley_values,)
        Shapley values.

    nbest : int, optional.
        Number of values to display on the bar plot. The nbest features with the highest Shapley values (absolute value)
        will be displayed. If the tolal number of regions is lower than nbest all the values will be displayed. Only
        valid when `sort` is True. The default is 10.

    names : list of strings or None, optional.
        Names of the regions associated to the Shapley values. If None default names 'instance_0', ..., 'instance_n'
        will be used. The default is None.

    sort : bool, optional.
        If true Shapley values are ranked from top to bottom in decreasing order with respect to their absolute value.
        If False all the regions will be displayed on the bar plot. The default is True.

    ax : matplotlib.axes, optional
            The default is None.

    Returns
    -------
    None.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 8))
    if names is not None:
        df = pd.DataFrame(shapley_values, index=names, columns=["shapley"])
    else:
        df = pd.DataFrame(
            shapley_values,
            index=["instance_" + str(i) for i in range(len(shapley_values))],
            columns=["shapley"],
        )

    df["shapley_abs"] = np.abs(df["shapley"])
    df["sign"] = 1 * (df["shapley"] > 0)
    if sort:
        df = df.sort_values(by="shapley_abs", ascending=False).iloc[
            : max(nbest, len(shapley_values)), :
        ]

    sns.barplot(
        data=df.reset_index(),
        orient="h",
        x="shapley",
        y="index",
        hue="sign",
        hue_order=[0, 1],
        palette=["blue", "red"],
        dodge=False,
        ax=ax,
    )
    ax.set(xlabel=None, ylabel=None)
    ax.set_axisbelow(True)
    ax.yaxis.grid(color="gray", linestyle="dashed")
    ax.legend().set_visible(False)
    ax.axvline(x=0, color="k", linestyle="--")
    # ax.set_xticks(
    #     ticks=ax.get_xticks(), labels=np.round(ax.get_xticks() + shap.empty_value, 4)
    # )
    sns.despine()
    return


def plot_pet(
    shapley_values: np.ndarray,
    image_path: str,
    masks_paths: list,
    save_path: Optional[Union[None, str]] = None,
    alpha: Optional[float] = 0.7,
    max_suv: Optional[float] = 10,
    cmap_name: Optional[str] = "seismic",
    cmap_lim: Optional[Union[None, float]] = None,
    plot_colorbar: Optional[bool] = True,
    title: Optional[Union[None, str]] = None,
) -> None:
    """Display the Shapley values of different instances on the Maxumum Intensity Projection (MIP) of a PET image.
    The different values are represented with a color scale.

    Parameters
    ----------
    shapley_values : 1D array shape (n_shapley_values,)
        Shapley values.

    image_path: string.
        Path of the PET image. The PET image format should be compatible with simpleitk. Nifti images (.nii.gz) are
        recommended

    masks_paths: list of strings.
        List of paths for the masks of the regions of interest associated with the Shapley values. Nifti images
        (.nii.gz) are recommended

    save_path: string or None, optional.
        Path to save the plot. If None the plotis not saved.
        The default is None.

    alpha: float in [0, 1], optional.
        Parameter to regulate the transparancy of the colored masks.
        The default is 0.7

    max_suv: float, optional
        Upper limit for the binary color mapping of the SUV intensities.
        The default is 10.

    cmap_name: string, optional
        The default is 'seismic'.

    cmap_lim: float or None, optional
        The default is None.

    plot_colorbar: bool, optional
        The default is True.

    title: string or None, optional
        Title of the plot. If None no title will be displayed.
        The default is None.

    Returns
    -------
    None.
    """
    # Load image and create MIP views
    img = sitk.ReadImage(image_path)
    img_array = sitk.GetArrayFromImage(img).astype(np.float32)  # [-450:, :, :]
    mip_1 = np.flipud(np.max(np.rot90(img_array, k=1, axes=(2, 1)), axis=-1))
    mip_2 = np.flipud(np.max(img_array, axis=-1))

    # shap_values = shap.shapleyvalues_.copy() + shap.empty_value
    # Define color map
    if cmap_lim is None:
        cmap = MplColorHelper(
            cmap_name=cmap_name,
            cmap_lim=(0, np.abs(shapley_values).max()),
            # cmap_lim=(shap.empty_value, np.abs(shap.shapleyvalues_).max()),
        )
    else:
        cmap = MplColorHelper(
            cmap_name=cmap_name,
            cmap_lim=(0, cmap_lim)
            # cmap_name=cmap_name, cmap_lim=(shap.empty_value, cmap_lim)
        )

    # Plot MIP views
    fig, axes = plt.subplots(1, 2, figsize=(16, 10))
    axes[0].imshow(mip_1, cmap="binary", vmin=0, vmax=max_suv)
    axes[1].imshow(mip_2, cmap="binary", vmin=0, vmax=max_suv)

    for mask_path, importance in zip(masks_paths, shapley_values):
        mask = sitk.ReadImage(mask_path)
        mask_array = sitk.GetArrayFromImage(mask).astype(np.float32)  # [-450:, :, :]
        mask_mip_1 = np.flipud(np.max(np.rot90(mask_array, k=1, axes=(2, 1)), axis=-1))
        mask_mip_2 = np.flipud(np.max(mask_array, axis=-1))
        color = cmap.get_rgb(importance, alpha=alpha)

        mask_1 = np.zeros((mask_mip_1.shape[0], mask_mip_1.shape[1], 4))
        mask_2 = np.zeros((mask_mip_2.shape[0], mask_mip_2.shape[1], 4))
        mask_1[mask_mip_1 == 1, :] = color
        mask_2[mask_mip_2 == 1, :] = color

        axes[0].imshow(mask_1)
        axes[1].imshow(mask_2)

    axes[0].axis("off")
    axes[1].axis("off")
    if plot_colorbar:
        fig.subplots_adjust(right=0.9)
        cbar_ax = fig.add_axes([0.85, 0.2, 0.02, 0.6])
        cb = fig.colorbar(mappable=cmap.scalarMap, aspect=80, cax=cbar_ax)
        cbar_ax.yaxis.set_label_position("left")
        cb.set_label("Shapley values", size=12, labelpad=5)
    if title is not None:
        fig.suptitle(title, fontsize=16, y=0.15)
    if save_path is not None:
        fig.savefig(save_path)
    return



from typing import NoReturn, Optional, Callable, Union

import numpy as np


class IndependentMaskedModel:
    """ Return a new model that masks out features associated with invalid inputs by integrating over a given
     background data set.

    Parameters
    ----------
    model: callable (input: 2D array of shape (n_inputs, n_input_features), output: 1D array of shape (n_inputs,))
        Original model whose invalid inputs need to be handled.

    background_data: 2D array of shape (n_background, n_input_features)
        Background data set used to deal with invalid inputs for the model. In the case of an invalid input, we
        generate "n_background" new inputs by replacing the invalid values with corresponding values from the background
        data set. We then apply the model to each of these generated inputs and return the average output.

    invalid_features: 1D array of shape (n_input_features)
        Boolean array to specify the features for which a NaN value means that the input is not valid. In such cases, we
        handle them with the background data set. If None, all the features will be taken into account to define invalid
        inputs.
    """

    def __init__(
            self,
            model: Callable[[np.ndarray], float],
            background_data: np.ndarray,
            invalid_features: Union[np.ndarray, None]
    ) -> NoReturn:

        self.model = model
        self.background_data = background_data
        if invalid_features is None:
            self.invalid_features = np.full(background_data.shape[1], True)
        else:
            self.invalid_features = invalid_features

    def __call__(self, X: np.ndarray) -> np.ndarray:
        nsamples = X.shape[0]
        results = np.zeros(nsamples)
        for i in range(nsamples):
            bool_mask = self.invalid_features * np.isnan(X[i, :])
            if np.any(bool_mask):
                Xnew = np.repeat(X[i, :].reshape(1, -1), repeats=self.background_data.shape[0], axis=0)
                Xnew[:, bool_mask] = self.background_data[:, bool_mask]
                results[i] = np.mean(self.model(Xnew))
            else:
                results[i] = self.model(X[i, :].reshape(1, -1))
        return results



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="radshap",
    version="0.0.1",
    author="Nicolas Captier",
    author_email="nicolas.captier@curie.fr",
    description="Shapley values for interpreting radiomic models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ncaptier/radshap",
    packages=setuptools.find_packages(),
    install_requires=[
        "matplotlib",
        "numpy",
        "joblib>=1.1.0",
        "pandas>=1.3.5",
        "seaborn>=0.11.2",
        "SimpleITK>=1.2.4"
    ],

    extras_require={"dev": ["pytest"],
                    "docs": ["sphinx==5.3.0", "sphinx-gallery == 0.14.0", "numpydoc == 1.5.0", "nbsphinx == 0.9.3",
                             "ipython==7.34.0", "sphinx-rtd-theme==1.3.0"]},

    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires='>=3.7',
)